@@ -5,7 +5,7 @@
 ;; Author: Daniel Mendler <mail@daniel-mendler.de>
 ;; Maintainer: Daniel Mendler <mail@daniel-mendler.de>
 ;; Created: 2022
-;; Version: 0.2
+;; Version: 0.3
 ;; Package-Requires: ((emacs "27.1"))
 ;; Homepage: https://github.com/minad/tempel
 
@@ -72,7 +72,7 @@ nil or a new template element, which is subsequently evaluated."
   :type 'hook)
 
 (defcustom tempel-template-sources
-  (list 'tempel-local-templates #'tempel--file-templates)
+  (list #'tempel-file-templates)
   "List of template sources.
 A source can either be a function or a variable symbol. The functions
 must return a list of templates which apply to the buffer or context."
@@ -102,9 +102,6 @@ must return a list of templates which apply to the buffer or context."
     (t :inherit highlight :slant italic))
   "Face used for default values.")
 
-(defvar-local tempel-local-templates nil
-  "List of templates which apply to the current buffer.")
-
 (defvar tempel--file-templates nil
   "Templates loaded from the `tempel-file'.")
 
@@ -114,11 +111,14 @@ must return a list of templates which apply to the buffer or context."
 (defvar tempel--history nil
   "Completion history used by `tempel-insert'.")
 
+(defvar tempel--inhibit-hooks nil
+  "Inhibit tempel modification change hooks from running.")
+
 (defvar-local tempel--active nil
   "List of active templates.
 Each template state is a pair, where the car is a list of overlays and
 the cdr is an alist of variable bindings. The template state is attached
-to each overlay as the property `tempel--state'. Furthermore overlays
+to each overlay as the property `tempel--field'. Furthermore overlays
 may be named with `tempel--name' or carry an evaluatable Lisp expression
 `tempel--form'.")
 
@@ -126,12 +126,17 @@ may be named with `tempel--name' or carry an evaluatable Lisp expression
   (let ((map (make-sparse-keymap)))
     (define-key map [remap beginning-of-buffer] #'tempel-beginning)
     (define-key map [remap end-of-buffer] #'tempel-end)
-    (define-key map [remap forward-paragraph] #'tempel-next)
-    (define-key map [remap backward-paragraph] #'tempel-previous)
-    (define-key map [remap keyboard-quit] #'tempel-abort)
+    (define-key map [remap kill-sentence] #'tempel-kill)
     (define-key map [remap keyboard-escape-quit] #'tempel-abort)
+    (define-key map [remap backward-paragraph] #'tempel-previous)
+    (define-key map [remap forward-paragraph] #'tempel-next)
+    ;; Use concrete keys because of org mode
+    (define-key map "\M-{" #'tempel-previous)
+    (define-key map "\M-}" #'tempel-next)
+    (define-key map [M-up] #'tempel-previous)
+    (define-key map [M-down] #'tempel-next)
     map)
-  "Keymap to navigate across template markers.")
+  "Keymap to navigate across template fields.")
 
 (defun tempel--print-element (elt)
   "Return string representation of template ELT."
@@ -175,38 +180,50 @@ REGION are the current region bouns"
       (tempel--delete-word name)
       (tempel--insert template region))))
 
+(defun tempel--range-modified (ov &rest _)
+  "Range overlay OV modified."
+  (when (and (not tempel--inhibit-hooks) (= (overlay-start ov) (overlay-end ov)))
+    (let ((st (overlay-get ov 'tempel--range)))
+      (setq tempel--active (cons st (delq st tempel--active)))
+      (tempel--disable))))
+
 (defun tempel--field-modified (ov after beg end &optional _len)
   "Update field overlay OV.
 AFTER is non-nil after the modification.
 BEG and END are the boundaries of the modification."
-  (cond
-   ;; Erase default before modification if at beginning or end
-   ((and (not after) (overlay-get ov 'tempel--default)
-         (or (= beg (overlay-start ov)) (= end (overlay-end ov))))
-    (delete-region (overlay-start ov) (overlay-end ov)))
-   ;; Update field after modification
-   (after
-    (let ((st (overlay-get ov 'tempel--state)))
-      (unless undo-in-progress
-        (move-overlay ov (overlay-start ov) (max end (overlay-end ov))))
-      (when-let (name (overlay-get ov 'tempel--name))
-        (setf (alist-get name (cdr st))
-              (buffer-substring-no-properties
-               (overlay-start ov) (overlay-end ov))))
-      (unless undo-in-progress
-        (tempel--synchronize-fields st ov)))))
-  (tempel--update-mark ov))
+  (unless tempel--inhibit-hooks
+    (cond
+     ;; Erase default before modification if at beginning or end
+     ((and (not after) (overlay-get ov 'tempel--default)
+           (or (= beg (overlay-start ov)) (= end (overlay-end ov))))
+      (delete-region (overlay-start ov) (overlay-end ov)))
+     ;; Update field after modification
+     (after
+      (let ((st (overlay-get ov 'tempel--field)))
+        (unless undo-in-progress
+          (move-overlay ov (overlay-start ov) (max end (overlay-end ov))))
+        (when-let (name (overlay-get ov 'tempel--name))
+          (setf (alist-get name (cdr st))
+                (buffer-substring-no-properties
+                 (overlay-start ov) (overlay-end ov))))
+        (unless undo-in-progress
+          (tempel--synchronize-fields st ov)))))
+    (tempel--update-mark ov)))
 
 (defun tempel--synchronize-fields (st current)
   "Synchronize fields of ST, except CURRENT overlay."
-  (dolist (ov (car st))
-    (unless (eq ov current)
-      (save-excursion
-        (goto-char (overlay-start ov))
-        (let (x)
-          (setq x (or (and (setq x (overlay-get ov 'tempel--form)) (eval x (cdr st)))
-                      (and (setq x (overlay-get ov 'tempel--name)) (alist-get x (cdr st)))))
-          (when x (tempel--replace (overlay-start ov) (overlay-end ov) ov x)))))))
+  (let ((range (caar st)))
+    (dolist (ov (cdar st))
+      (unless (eq ov current)
+        (save-excursion
+          (goto-char (overlay-start ov))
+          (let (x)
+            (setq x (or (and (setq x (overlay-get ov 'tempel--form)) (eval x (cdr st)))
+                        (and (setq x (overlay-get ov 'tempel--name)) (alist-get x (cdr st)))))
+            (when x (tempel--replace (overlay-start ov) (overlay-end ov) ov x)))))
+      ;; Move range overlay
+      (move-overlay range (overlay-start range)
+                    (max (overlay-end range) (overlay-end ov))))))
 
 (defun tempel--replace (beg end ov str)
   "Replace region beween BEG and END with STR.
@@ -217,7 +234,8 @@ If OV is alive, move it."
       (unless (eq buffer-undo-list t)
         (push (list 'apply #'tempel--replace beg (+ beg (length str)) ov old)
               buffer-undo-list))
-      (with-silent-modifications
+      (let ((buffer-undo-list t)
+            (tempel--inhibit-hooks t))
         (save-excursion
           (goto-char beg)
           (delete-char (- end beg))
@@ -250,7 +268,7 @@ INIT is the optional initial input."
       (insert init)
       (move-overlay ov (overlay-start ov) (point)))
     (tempel--update-mark ov)
-    (overlay-put ov 'tempel--state st)
+    (overlay-put ov 'tempel--field st)
     (overlay-put ov 'modification-hooks (list #'tempel--field-modified))
     (overlay-put ov 'insert-in-front-hooks (list #'tempel--field-modified))
     (overlay-put ov 'insert-behind-hooks (list #'tempel--field-modified))
@@ -286,7 +304,7 @@ INIT is the optional initial input."
     ('% (unless (or (eolp) (save-excursion (re-search-forward "\\=\\s-*$" nil t)))
           (insert "\n")))
     ('o (unless (or (eolp) (save-excursion (re-search-forward "\\=\\s-*$" nil t)))
-	  (open-line 1)))
+          (open-line 1)))
     (`(s ,name) (tempel--field st name))
     (`(l . ,lst) (dolist (e lst) (tempel--element st region e)))
     ((or 'p `(,(or 'p 'P) . ,rest)) (apply #'tempel--placeholder st rest))
@@ -316,35 +334,43 @@ PROMPT is the optional prompt/default value."
 
 (defun tempel--insert (template region)
   "Insert TEMPLATE given the current REGION."
-  ;; TODO do we want to have the ability to reactivate snippets?
-  (unless (eq buffer-undo-list t)
-    (push (list 'apply #'tempel--disable) buffer-undo-list))
-  (setf (alist-get 'tempel--active minor-mode-overriding-map-alist) tempel-map)
-  (save-excursion
-    ;; Split existing overlays, do not expand within existing field.
-    ;; TODO This will be causing issues. Think more about nested expansion.
-    (dolist (st tempel--active)
-      (dolist (ov (car st))
-        (when (and (<= (overlay-start ov) (point)) (>= (overlay-end ov) (point)))
-          (setf (overlay-end ov) (point)))))
-    ;; Activate template
-    (let ((st (cons nil nil))
-          (inhibit-modification-hooks t))
-      (push (make-overlay (point) (point)) (car st))
-      (overlay-put (caar st) 'face 'cursor) ;; TODO debug
-      (dolist (elt template) (tempel--element st region elt))
-      (push (make-overlay (point) (point) nil t t) (car st))
-      (overlay-put (caar st) 'face 'cursor) ;; TODO debug
-      (push st tempel--active)))
-  (if (cddaar tempel--active)
-      (unless (cl-loop for ov in (caar tempel--active)
-                       thereis (and (overlay-get ov 'tempel--state)
-                                    (eq (point) (overlay-start ov))))
-        ;; Jump to first field
-        (tempel-next 1))
-    ;; Disable right away
-    (goto-char (overlay-start (caaar tempel--active)))
-    (tempel--disable)))
+  (let ((plist template))
+    (while (and plist (not (keywordp (car plist))))
+      (pop plist))
+    (eval (plist-get plist :pre) 'lexical)
+    ;; TODO do we want to have the ability to reactivate snippets?
+    (unless (eq buffer-undo-list t)
+      (push (list 'apply #'tempel--disable) buffer-undo-list))
+    (setf (alist-get 'tempel--active minor-mode-overriding-map-alist) tempel-map)
+    (save-excursion
+      ;; Split existing overlays, do not expand within existing field.
+      ;; TODO This will be causing issues. Think more about nested expansion.
+      (dolist (st tempel--active)
+        (dolist (ov (cdar st))
+          (when (and (<= (overlay-start ov) (point)) (>= (overlay-end ov) (point)))
+            (setf (overlay-end ov) (point)))))
+      ;; Activate template
+      (let ((st (cons nil nil))
+            (range (point))
+            (tempel--inhibit-hooks t))
+        (while (and template (not (keywordp (car template))))
+          (tempel--element st region (pop template)))
+        (setq range (make-overlay range (point) nil t))
+        (push range (car st))
+        (overlay-put range 'modification-hooks (list #'tempel--range-modified))
+        (overlay-put range 'tempel--range st)
+        ;;(overlay-put range 'face 'region) ;; TODO debug
+        (push st tempel--active)))
+    (cond
+     ((cl-loop for ov in (caar tempel--active)
+               never (overlay-get ov 'tempel--field))
+      (goto-char (overlay-end (caaar tempel--active)))
+      (tempel--disable)) ;; Disable right away
+     ((cl-loop for ov in (caar tempel--active)
+               never (and (overlay-get ov 'tempel--field)
+                          (eq (point) (overlay-start ov))))
+      (tempel-next 1))) ;; Jump to first field
+    (eval (plist-get plist :post) 'lexical)))
 
 (defun tempel--save ()
   "Save template file buffer."
@@ -361,25 +387,38 @@ PROMPT is the optional prompt/default value."
     (goto-char (point-max))
     (insert "\n)")
     (goto-char (point-min))
-    (let ((templates (read (current-buffer))) result)
-      (while (and templates (symbolp (car templates)))
-        (let ((mode (pop templates)) list)
-          (while (and templates (consp (car templates)))
-            (push (pop templates) list))
-          (push (cons mode (nreverse list)) result)))
+    (let ((data (read (current-buffer))) result)
+      (while data
+        (let (modes plist templates)
+          (while (and (car data) (symbolp (car data)) (not (keywordp (car data))))
+            (push (pop data) modes))
+          (while (keywordp (car data))
+            (push (pop data) plist)
+            (push (pop data) plist))
+          (while (consp (car data))
+            (push (pop data) templates))
+          (push `(,(nreverse modes) ,(nreverse plist) . ,(nreverse templates)) result)))
       result)))
 
-(defun tempel--file-templates ()
-  "Return templates defined in `tempel-file'."
+(defun tempel-file-templates ()
+  "Load the templates defined in `tempel-file'."
   (let ((mod (time-convert (file-attribute-modification-time
                             (file-attributes tempel-file))
                            'integer)))
     (unless (equal tempel--file-modified mod)
       (setq tempel--file-templates (tempel--file-read tempel-file)
             tempel--file-modified mod)))
-  (cl-loop for x in tempel--file-templates
-           if (or (derived-mode-p (car x)) (eq (car x) 'fundamental-mode))
-           append (cdr x)))
+  (cl-loop
+   for (modes plist . templates) in tempel--file-templates
+   if (and
+       (cl-loop for m in modes
+                thereis (or (derived-mode-p m) (eq m #'fundamental-mode)))
+       (or (not (plist-member plist :condition))
+           (save-excursion
+             (save-restriction
+               (save-match-data
+                 (eval (plist-get plist :condition) 'lexical))))))
+   append templates))
 
 (defun tempel--templates ()
   "Return templates for current mode."
@@ -427,6 +466,19 @@ PROMPT is the optional prompt/default value."
   (when-let (pos (tempel--end))
     (if (= pos (point)) (tempel-done) (goto-char pos))))
 
+(defun tempel--field-at-point ()
+  "Return the field overlay at point."
+  (cl-loop for ov in (overlays-in (max (point-min) (1- (point)))
+                                  (min (point-max) (1+ (point))))
+           thereis (and (overlay-get ov 'tempel--field) ov)))
+
+(defun tempel-kill ()
+  "Kill the field contents."
+  (interactive)
+  (if-let (ov (tempel--field-at-point))
+      (kill-region (overlay-start ov) (overlay-end ov))
+    (kill-sentence nil)))
+
 (defun tempel-next (arg)
   "Move ARG fields forward and quit at the end."
   (interactive "p")
@@ -443,14 +495,12 @@ PROMPT is the optional prompt/default value."
 (defun tempel--beginning ()
   "Return beginning of template markers."
   (and tempel--active
-       (cl-loop for st in tempel--active minimize
-                (cl-loop for ov in (car st) minimize (overlay-start ov)))))
+       (cl-loop for st in tempel--active minimize (overlay-start (caar st)))))
 
 (defun tempel--end ()
   "Return end of template markers."
   (and tempel--active
-       (cl-loop for st in tempel--active maximize
-                (cl-loop for ov in (car st) maximize (overlay-end ov)))))
+       (cl-loop for st in tempel--active maximize (overlay-end (caar st)))))
 
 (defun tempel-abort ()
   "Abort template insertion."
@@ -604,5 +654,14 @@ If called interactively, select a template with `completing-read'."
   (unless (or noninteractive (eq (aref (buffer-name) 0) ?\s))
     (tempel-abbrev-mode 1)))
 
+;; Emacs 28: Do not show Tempel commands in M-X
+(dolist (sym (list #'tempel-next #'tempel-previous #'tempel-beginning
+                   #'tempel-end #'tempel-kill #'tempel-done #'tempel-abort))
+  (put sym 'completion-predicate #'tempel--command-p))
+
+(defun tempel--command-p (_sym buffer)
+  "Return non-nil if Tempel is active in BUFFER."
+  (buffer-local-value 'tempel--active buffer))
+
 (provide 'tempel)
 ;;; tempel.el ends here
